@@ -1,4 +1,4 @@
-# Copyright (c) 2008-2009 Twisted Matrix Laboratories.
+# Copyright (c) Twisted Matrix Laboratories.
 # See LICENSE for details.
 
 """
@@ -7,36 +7,36 @@ Tests for implementations of L{IReactorProcess}.
 
 __metaclass__ = type
 
-import os
-import warnings, sys, signal
+import os, sys, signal, threading
 
+from twisted.trial.unittest import TestCase, SkipTest
 from twisted.internet.test.reactormixins import ReactorBuilder
-from twisted.python.compat import set
 from twisted.python.log import msg, err
-from twisted.python.runtime import platform
+from twisted.python.runtime import platform, platformType
 from twisted.python.filepath import FilePath
-from twisted.python.failure import Failure
 from twisted.internet import utils
+from twisted.internet.interfaces import IReactorProcess, IProcessTransport
 from twisted.internet.defer import Deferred, succeed
 from twisted.internet.protocol import ProcessProtocol
-from twisted.internet.error import ProcessDone, PotentialZombieWarning
-from twisted.internet.error import ProcessTerminated
+from twisted.internet.error import ProcessDone, ProcessTerminated
+
 
-skipWindowsNopywin32 = None
-if platform.isWindows():
-    try:
-        import win32process
-    except ImportError:
-        skipWindowsNopywin32 = ("On windows, spawnProcess is not available "
-                                "in the absence of win32process.")
 
 class _ShutdownCallbackProcessProtocol(ProcessProtocol):
     """
     An L{IProcessProtocol} which fires a Deferred when the process it is
     associated with ends.
+
+    @ivar received: A C{dict} mapping file descriptors to lists of bytes
+        received from the child process on those file descriptors.
     """
     def __init__(self, whenFinished):
         self.whenFinished = whenFinished
+        self.received = {}
+
+
+    def childDataReceived(self, fd, bytes):
+        self.received.setdefault(fd, []).append(bytes)
 
 
     def processEnded(self, reason):
@@ -45,75 +45,181 @@ class _ShutdownCallbackProcessProtocol(ProcessProtocol):
 
 
 class ProcessTestsBuilderBase(ReactorBuilder):
-    def spawnProcess(self, reactor):
-        """
-        Call C{reactor.spawnProcess} with some simple arguments.  Do this here
-        so that code object referenced by the stack frame has a C{co_filename}
-        attribute set to this file so that L{TestCase.assertWarns} can be used.
+    """
+    Base class for L{IReactorProcess} tests which defines some tests which
+    can be applied to PTY or non-PTY uses of C{spawnProcess}.
 
-        Return a L{Deferred} which fires when the process has exited.
+    Subclasses are expected to set the C{usePTY} attribute to C{True} or
+    C{False}.
+    """
+    requiredInterfaces = [IReactorProcess]
+
+
+    def test_processTransportInterface(self):
         """
-        onConnection = Deferred()
+        L{IReactorProcess.spawnProcess} connects the protocol passed to it
+        to a transport which provides L{IProcessTransport}.
+        """
+        ended = Deferred()
+        protocol = _ShutdownCallbackProcessProtocol(ended)
+
+        reactor = self.buildReactor()
+        transport = reactor.spawnProcess(
+            protocol, sys.executable, [sys.executable, "-c", ""],
+            usePTY=self.usePTY)
+
+        # The transport is available synchronously, so we can check it right
+        # away (unlike many transport-based tests).  This is convenient even
+        # though it's probably not how the spawnProcess interface should really
+        # work.
+        # We're not using verifyObject here because part of
+        # IProcessTransport is a lie - there are no getHost or getPeer
+        # methods.  See #1124.
+        self.assertTrue(IProcessTransport.providedBy(transport))
+
+        # Let the process run and exit so we don't leave a zombie around.
+        ended.addCallback(lambda ignored: reactor.stop())
+        self.runReactor(reactor)
+
+
+    def _writeTest(self, write):
+        """
+        Helper for testing L{IProcessTransport} write functionality.  This
+        method spawns a child process and gives C{write} a chance to write some
+        bytes to it.  It then verifies that the bytes were actually written to
+        it (by relying on the child process to echo them back).
+
+        @param write: A two-argument callable.  This is invoked with a process
+            transport and some bytes to write to it.
+        """
+        reactor = self.buildReactor()
+
+        ended = Deferred()
+        protocol = _ShutdownCallbackProcessProtocol(ended)
+
+        bytes = "hello, world" + os.linesep
+        program = (
+            "import sys\n"
+            "sys.stdout.write(sys.stdin.readline())\n"
+            )
+
+        def startup():
+            transport = reactor.spawnProcess(
+                protocol, sys.executable, [sys.executable, "-c", program])
+            try:
+                write(transport, bytes)
+            except:
+                err(None, "Unhandled exception while writing")
+                transport.signalProcess('KILL')
+        reactor.callWhenRunning(startup)
+
+        ended.addCallback(lambda ignored: reactor.stop())
+
+        self.runReactor(reactor)
+        self.assertEqual(bytes, "".join(protocol.received[1]))
+
+
+    def test_write(self):
+        """
+        L{IProcessTransport.write} writes the specified C{str} to the standard
+        input of the child process.
+        """
+        def write(transport, bytes):
+            transport.write(bytes)
+        self._writeTest(write)
+
+
+    def test_writeSequence(self):
+        """
+        L{IProcessTransport.writeSequence} writes the specified C{list} of
+        C{str} to the standard input of the child process.
+        """
+        def write(transport, bytes):
+            transport.writeSequence(list(bytes))
+        self._writeTest(write)
+
+
+    def test_writeToChild(self):
+        """
+        L{IProcessTransport.writeToChild} writes the specified C{str} to the
+        specified file descriptor of the child process.
+        """
+        def write(transport, bytes):
+            transport.writeToChild(0, bytes)
+        self._writeTest(write)
+
+
+    def test_writeToChildBadFileDescriptor(self):
+        """
+        L{IProcessTransport.writeToChild} raises L{KeyError} if passed a file
+        descriptor which is was not set up by L{IReactorProcess.spawnProcess}.
+        """
+        def write(transport, bytes):
+            try:
+                self.assertRaises(KeyError, transport.writeToChild, 13, bytes)
+            finally:
+                # Just get the process to exit so the test can complete
+                transport.write(bytes)
+        self._writeTest(write)
+
+
+    def test_spawnProcessEarlyIsReaped(self):
+        """
+        If, before the reactor is started with L{IReactorCore.run}, a
+        process is started with L{IReactorProcess.spawnProcess} and
+        terminates, the process is reaped once the reactor is started.
+        """
+        reactor = self.buildReactor()
+
+        # Create the process with no shared file descriptors, so that there
+        # are no other events for the reactor to notice and "cheat" with.
+        # We want to be sure it's really dealing with the process exiting,
+        # not some associated event.
+        if self.usePTY:
+            childFDs = None
+        else:
+            childFDs = {}
+
+        # Arrange to notice the SIGCHLD.
+        signaled = threading.Event()
+        def handler(*args):
+            signaled.set()
+        signal.signal(signal.SIGCHLD, handler)
+
+        # Start a process - before starting the reactor!
+        ended = Deferred()
         reactor.spawnProcess(
-            _ShutdownCallbackProcessProtocol(onConnection), sys.executable,
-            [sys.executable, "-c", ""], usePTY=self.usePTY)
-        return onConnection
+            _ShutdownCallbackProcessProtocol(ended), sys.executable,
+            [sys.executable, "-c", ""], usePTY=self.usePTY, childFDs=childFDs)
 
+        # Wait for the SIGCHLD (which might have been delivered before we got
+        # here, but that's okay because the signal handler was installed above,
+        # before we could have gotten it).
+        signaled.wait(120)
+        if not signaled.isSet():
+            self.fail("Timed out waiting for child process to exit.")
 
-    def test_spawnProcessTooEarlyWarns(self):
-        """
-        C{reactor.spawnProcess} emits a warning if it is called before
-        C{reactor.run}.
+        # Capture the processEnded callback.
+        result = []
+        ended.addCallback(result.append)
 
-        If you can figure out a way to make it safe to run
-        C{reactor.spawnProcess} before C{reactor.run}, you may delete the
-        warning and this test.
-        """
-        reactor = self.buildReactor()
-        whenFinished = self.assertWarns(
-            PotentialZombieWarning,
-            PotentialZombieWarning.MESSAGE, __file__,
-            self.spawnProcess, reactor)
-
-        def check():
-            # Handle the exact problem the warning we're testing for is about.
-            # If the SIGCHLD arrives before we install our SIGCHLD handler,
-            # we'll never see the process exit.  So after the SIGCHLD handler
-            # is installed, try to reap children, just in case.
-            from twisted.internet.process import reapAllProcesses
-            reapAllProcesses()
-        reactor.callWhenRunning(check)
-
-        # Now wait for the process to exit before finishing the test.  If we
-        # don't do this, tearDown might close the PTY before the child is done
-        # with it.
-        whenFinished.addCallback(
-            lambda ignored: reactor.callWhenRunning(reactor.stop))
+        if result:
+            # The synchronous path through spawnProcess / Process.__init__ /
+            # registerReapProcessHandler was encountered.  There's no reason to
+            # start the reactor, because everything is done already.
+            return
 
+        # Otherwise, though, start the reactor so it can tell us the process
+        # exited.
+        ended.addCallback(lambda ignored: reactor.stop())
         self.runReactor(reactor)
 
-
-    def test_callWhenRunningSpawnProcessWarningFree(self):
-        """
-        L{PotentialZombieWarning} is not emitted when the reactor is run after
-        C{reactor.callWhenRunning(reactor.spawnProcess, ...)} has been called.
-        """
-        reactor = self.buildReactor()
-        def spawnProcess():
-            whenFinished = self.spawnProcess(reactor)
-            # Wait for the process to exit before finishing the test.  If we
-            # don't do this, tearDown might close the PTY before the child is
-            # done with it.
-            whenFinished.addCallback(lambda ign: reactor.stop())
-        reactor.callWhenRunning(spawnProcess)
-        self.runReactor(reactor)
-        self.assertEqual(self.flushWarnings(), [])
-
+        # Make sure the reactor stopped because the Deferred fired.
+        self.assertTrue(result)
 
     if getattr(signal, 'SIGCHLD', None) is None:
-        skipMsg = "No SIGCHLD, no zombies possible."
-        test_spawnProcessTooEarlyWarns.skip = skipMsg
-        test_callWhenRunningSpawnProcessWarningFree.skip = skipMsg
+        test_spawnProcessEarlyIsReaped.skip = (
+            "Platform lacks SIGCHLD, early-spawnProcess test can't work.")
 
 
     def test_processExitedWithSignal(self):
@@ -185,6 +291,108 @@ class ProcessTestsBuilderBase(ReactorBuilder):
         self.runReactor(reactor)
 
 
+    def test_systemCallUninterruptedByChildExit(self):
+        """
+        If a child process exits while a system call is in progress, the system
+        call should not be interfered with.  In particular, it should not fail
+        with EINTR.
+
+        Older versions of Twisted installed a SIGCHLD handler on POSIX without
+        using the feature exposed by the SA_RESTART flag to sigaction(2).  The
+        most noticable problem this caused was for blocking reads and writes to
+        sometimes fail with EINTR.
+        """
+        reactor = self.buildReactor()
+        result = []
+
+        def f():
+            try:
+                f1 = os.popen('%s -c "import time; time.sleep(0.1)"' %
+                    (sys.executable,))
+                f2 = os.popen('%s -c "import time; time.sleep(0.5); print \'Foo\'"' %
+                    (sys.executable,))
+                # The read call below will blow up with an EINTR from the
+                # SIGCHLD from the first process exiting if we install a
+                # SIGCHLD handler without SA_RESTART.  (which we used to do)
+                result.append(f2.read())
+            finally:
+                reactor.stop()
+
+        reactor.callWhenRunning(f)
+        self.runReactor(reactor)
+        self.assertEqual(result, ["Foo\n"])
+
+
+    def test_openFileDescriptors(self):
+        """
+        A spawned process has only stdin, stdout and stderr open
+        (file descriptor 3 is also reported as open, because of the call to
+        'os.listdir()').
+        """
+        here = FilePath(__file__)
+        top = here.parent().parent().parent().parent()
+        source = (
+            "import sys",
+            "sys.path.insert(0, '%s')" % (top.path,),
+            "from twisted.internet import process",
+            "sys.stdout.write(str(process._listOpenFDs()))",
+            "sys.stdout.flush()")
+
+        def checkOutput(output):
+            self.assertEqual('[0, 1, 2, 3]', output)
+
+        reactor = self.buildReactor()
+
+        class Protocol(ProcessProtocol):
+            def __init__(self):
+                self.output = []
+
+            def outReceived(self, data):
+                self.output.append(data)
+
+            def processEnded(self, reason):
+                try:
+                    checkOutput("".join(self.output))
+                finally:
+                    reactor.stop()
+
+        proto = Protocol()
+        reactor.callWhenRunning(
+            reactor.spawnProcess, proto, sys.executable,
+            [sys.executable, "-Wignore", "-c", "\n".join(source)],
+            usePTY=self.usePTY)
+        self.runReactor(reactor)
+
+    if platformType != "posix":
+        test_openFileDescriptors.skip = "Test only applies to POSIX platforms"
+
+
+    def test_timelyProcessExited(self):
+        """
+        If a spawned process exits, C{processExited} will be called in a
+        timely manner.
+        """
+        reactor = self.buildReactor()
+
+        class ExitingProtocol(ProcessProtocol):
+            exited = False
+
+            def processExited(protoSelf, reason):
+                protoSelf.exited = True
+                reactor.stop()
+                self.assertEqual(reason.value.exitCode, 0)
+
+        protocol = ExitingProtocol()
+        reactor.callWhenRunning(
+            reactor.spawnProcess, protocol, sys.executable,
+            [sys.executable, "-c", "raise SystemExit(0)"],
+            usePTY=self.usePTY)
+
+        # This will timeout if processExited isn't called:
+        self.runReactor(reactor, timeout=30)
+        self.assertEqual(protocol.exited, True)
+
+
 
 class ProcessTestsBuilder(ProcessTestsBuilderBase):
     """
@@ -427,7 +635,7 @@ class ProcessTestsBuilder(ProcessTestsBuilderBase):
             output = output.split('\0')
             # Drop the trailing \0.
             output.pop()
-            self.assertEquals(args, output)
+            self.assertEqual(args, output)
 
         def shutdown(result):
             reactor.stop()
@@ -442,10 +650,6 @@ class ProcessTestsBuilder(ProcessTestsBuilderBase):
 
         reactor.callWhenRunning(spawnChild)
         self.runReactor(reactor)
-
-
-
-ProcessTestsBuilder.skip = skipWindowsNopywin32
 globals().update(ProcessTestsBuilder.makeTestCaseClasses())
 
 
@@ -463,6 +667,28 @@ class PTYProcessTestsBuilder(ProcessTestsBuilderBase):
         skippedReactors = {
             "twisted.internet.pollreactor.PollReactor":
                 "OS X's poll() does not support PTYs"}
-
-
 globals().update(PTYProcessTestsBuilder.makeTestCaseClasses())
+
+
+
+class PotentialZombieWarningTests(TestCase):
+    """
+    Tests for L{twisted.internet.error.PotentialZombieWarning}.
+    """
+    def test_deprecated(self):
+        """
+        Accessing L{PotentialZombieWarning} via the
+        I{PotentialZombieWarning} attribute of L{twisted.internet.error}
+        results in a deprecation warning being emitted.
+        """
+        from twisted.internet import error
+        error.PotentialZombieWarning
+
+        warnings = self.flushWarnings([self.test_deprecated])
+        self.assertEqual(warnings[0]['category'], DeprecationWarning)
+        self.assertEqual(
+            warnings[0]['message'],
+            "twisted.internet.error.PotentialZombieWarning was deprecated in "
+            "Twisted 10.0.0: There is no longer any potential for zombie "
+            "process.")
+        self.assertEqual(len(warnings), 1)
